from tests import BaseTest

from pymatex.node import *


class OtherTests(BaseTest.BaseTest):

    def test_read_zeta(self):
        ast = self.parse(r'\frac{1}{1 - 2^{1-s}} \sum_{n=1}^{\infty} \frac{(-1)^{n+1}}{n^s}')
        one = Constant('1')
        s = Variable('s')
        n = Variable('n')
        self.assertEqual(ast, Multiplication(
            Division(one, Subtraction(one, Exponentiation(Constant('2'), Subtraction(one, s)))),
            Summation(n, one, Constant('\\infty'),
                      Division(Exponentiation(Negate(one), Addition(n, one)), Exponentiation(n, s)))))

    def test_read_multi_multiplication_of_expression(self):
        ast = self.parse(r'x y \sum_{n=1}^{\infty} n')
        n = Variable('n')
        self.assertEqual(ast, Multiplication(Multiplication(Variable('x'), Variable('y')), Summation(n, Constant('1'), Constant('\\infty'), n)))

    def test_read_multiplication_of_function(self):
        ast = self.parse(r'1/(n + 1/2) \sum_{k=1}^{n-1} \zeta(2k) \zeta(2n - 2k)')
        one = Constant('1')
        two = Constant('2')
        k = Variable('k')
        n = Variable('n')
        two_k = Multiplication(two, k)
        self.assertEqual(ast, Multiplication(Division(one, Addition(n, Division(one, two))),
                                             Summation(k, one, Subtraction(n, one),
                                                       Multiplication(Function(Func.ZETA, two_k), Function(Func.ZETA,
                                                                                                           Subtraction(
                                                                                                               Multiplication(
                                                                                                                   two,
                                                                                                                   n),
                                                                                                               two_k))))))

    def test_read_neville_theta_function(self):
        ast = self.parse(r'\frac{\sqrt{2\pi} q(m)^{1/4}}{m^{1/4} \sqrt{K(m)}} \sum_{k=0}^{\infty} (q(m))^{k (k+1)} \cos{\frac{(2k+1)\pi z}{2 K(m)}}')
        one = Constant('1')
        two = Constant('2')
        k = Variable('k')
        m = Variable('m')
        pi = Variable('\\pi')
        one_fourth = Division(one, Constant('4'))

        first_division = Division(Multiplication(Function(Func.SQRT, Multiplication(two, pi)), Exponentiation(Function('q', m), one_fourth)),
                                  Multiplication(Exponentiation(m, one_fourth), Function(Func.SQRT, Function('K', m))))
        q_m_exponent = Exponentiation(Function('q', m), Multiplication(k, Addition(k, one)))
        cos_thing = Function(Func.COS, Division(Multiplication(Multiplication(Addition(Multiplication(two, k), one), pi), Variable('z')), Multiplication(two, Function('K', m))))
        summation = Summation(k, Constant('0'), Constant('\\infty'), Multiplication(q_m_exponent, cos_thing))
        self.assertEqual(ast, Multiplication(first_division, summation))



from antlr4 import CommonTokenStream, InputStream, ParseTreeWalker
from pymatex.search import IndexCreatorVisitor, IndexSearchVisitor
from pymatex.listener import MatexAST
from pymatex.grammar import MatexLexer, MatexParser
import operator


class SearchQuery:

    def __init__(self):

        self.data = {}
        self.number_of_different_nodes = {}

    def __init__(self, path: str):

        self.data = {}
        self.number_of_different_nodes = {}

        with open(path, 'r') as f:
            for line in f:
                tokens = line.split(':')
                pk = int(tokens[0])
                latex = tokens[1].strip()

                ast = self.__parse(latex)

                visitor = IndexCreatorVisitor(self.data, pk)
                ast.accept(visitor)

                self.number_of_different_nodes[pk] = visitor.get_number_of_nodes_of_different_nodes()

    def add(self, pk: int, latex: str):
        ast = self.__parse(latex)

        visitor = IndexCreatorVisitor(self.data, pk)
        ast.accept(visitor)

        self.number_of_different_nodes[pk] = visitor.get_number_of_nodes_of_different_nodes()

    def is_valid(self, content: str):
        MatexAST.get_tree(content)

    def remove(self, pk: int):
        self.number_of_different_nodes.pop(pk)
        self.__internal_remove(self.data, pk)

    def search(self, content: str):
        ast = self.__parse(content)

        visitor = IndexSearchVisitor(self.data)
        ast.accept(visitor)

        results = visitor.get_results()

        for result in results:
            results[result] = int(results[result] / self.number_of_different_nodes[result])

        sorted_result = sorted(results.items(), key=operator.itemgetter(1), reverse=True)
        return sorted_result

    def __internal_remove(self, data, pk: int):
        if isinstance(data, dict):
            for k, v in data.items():
                if isinstance(v, dict) or isinstance(v, set):
                    self.__internal_remove(v, pk)
                else:
                    if v == pk:
                        del data[k]
        elif isinstance(data, set):
            data.discard(pk)

    def __parse(self, content: str):
        return MatexAST.parse(content)



from pymatex.listener import MatexASTVisitor
from pymatex.node.Node import Node


class Variable(Node):

    def __init__(self, variable: str):
        super().__init__()

        self.variable = variable

    def __eq__(self, other):
        return self.variable == other.variable

    def __str__(self):
        return self.variable

    def accept(self, visitor: MatexASTVisitor):
        return visitor.visit_variable(self)



from enum import Enum, auto
from pymatex.node.Node import Node


class IterativeType(Enum):
    INTEGRAL = auto()
    PRODUCT = auto()
    SUM = auto()


class IterativeFunction(Node):

    def __init__(self, iterative_type: IterativeType, variable, start_range, end_range, expression):
        super().__init__()

        self.iterative_type = iterative_type
        self.variable = variable
        self.start_range = start_range
        self.end_range = end_range
        self.expression = expression

    def __eq__(self, other):
        return self.iterative_type == other.iterative_type and \
               self.variable == other.variable and self.start_range == other.start_range and \
               self.end_range == other.end_range and self.expression == other.expression



from tests import BaseTest
from pymatex.node import NodeType
from pymatex.search import IndexCreatorVisitor


class IndexCreatorVisitorTests(BaseTest.BaseTest):

    def test_visit_basic_addition(self):
        ast = self.parse(r'3 + x')
        data = {}
        pk = 1
        visitor = IndexCreatorVisitor(data, pk)
        ast.accept(visitor)

        self.assertTrue(self.has(data[0][NodeType.CONSTANT], pk, '3'))
        self.assertTrue(self.has(data[0][NodeType.VARIABLE], pk, 'x'))
        self.assertTrue(pk in data[1][NodeType.ADDITION])

    def test_visit_addition_mixed_with_multiplication(self):
        ast = self.parse(r'3 * x + y * 5')
        data = {}
        pk = 1
        visitor = IndexCreatorVisitor(data, pk)
        ast.accept(visitor)

        self.assertTrue(self.has(data[0][NodeType.CONSTANT], pk, '3'))
        self.assertTrue(self.has(data[0][NodeType.VARIABLE], pk, 'x'))
        self.assertTrue(self.has(data[0][NodeType.VARIABLE], pk, 'y'))
        self.assertTrue(self.has(data[0][NodeType.CONSTANT], pk, '5'))
        self.assertTrue(pk in data[1][NodeType.MULTIPLICATION])
        self.assertTrue(pk in data[2][NodeType.ADDITION])

    def test_visit_summation(self):
        ast = self.parse(r'\sum_{i=0}^{\infty} i * i')
        data = {}
        pk = 1
        visitor = IndexCreatorVisitor(data, pk)
        ast.accept(visitor)

        self.assertTrue(self.has(data[0][NodeType.CONSTANT], pk, '0'))
        self.assertTrue(self.has(data[0][NodeType.BOUNDVARIABLE], pk, 'i'))
        self.assertTrue(self.has(data[0][NodeType.CONSTANT], pk, '\\infty'))
        self.assertTrue(pk in data[1][NodeType.MULTIPLICATION])
        self.assertTrue(pk in data[2][NodeType.SUMMATION])

    def test_visit_summation_with_free_and_bound_variable(self):
        ast = self.parse(r'\sum_{i=0}^{\infty} (i + j)')
        data = {}
        pk = 1
        visitor = IndexCreatorVisitor(data, pk)
        ast.accept(visitor)

        self.assertTrue(self.has(data[0][NodeType.CONSTANT], pk, '0'))
        self.assertTrue(self.has(data[0][NodeType.BOUNDVARIABLE], pk, 'i'))
        self.assertTrue(self.has(data[0][NodeType.VARIABLE], pk, 'j'))
        self.assertTrue(self.has(data[0][NodeType.CONSTANT], pk, '\\infty'))
        self.assertTrue(pk in data[1][NodeType.ADDITION])
        self.assertTrue(pk in data[2][NodeType.SUMMATION])

    def test_visit_exponentiation_and_function(self):
        ast = self.parse(r'e^{x!}')
        data = {}
        pk = 1
        visitor = IndexCreatorVisitor(data, pk)
        ast.accept(visitor)

        self.assertTrue(self.has(data[0][NodeType.VARIABLE], pk, 'e'))
        self.assertTrue(self.has(data[0][NodeType.VARIABLE], pk, 'x'))
        self.assertTrue(pk in data[1][NodeType.FUNCTION])
        self.assertTrue(pk in data[2][NodeType.EXPONENTIATION])

    def test_visit_indexed_variable(self):
        ast = self.parse(r'3*B_{2n}')
        data = {}
        pk = 1
        visitor = IndexCreatorVisitor(data, pk)
        ast.accept(visitor)

        self.assertTrue(self.has(data[0][NodeType.CONSTANT], pk, '2'))
        self.assertTrue(self.has(data[0][NodeType.CONSTANT], pk, '3'))
        self.assertTrue(self.has(data[0][NodeType.VARIABLE], pk, 'n'))
        self.assertTrue(pk in data[1][NodeType.MULTIPLICATION])
        self.assertTrue(self.has(data[2][NodeType.INDEXEDVARIABLE], pk, 'B'))
        self.assertTrue(pk in data[3][NodeType.MULTIPLICATION])

    def has(self, data: dict, pk: int, constant_value: str):
        self.assertIn(constant_value, data)
        self.assertIn(pk, data[constant_value])
        return True



from tests import BaseTest
from pymatex.search import SearchQuery


class SearchQueryTests(BaseTest.BaseTest):

    def test_read(self):
        s = SearchQuery(path='tests/search/resources/search-content-simple-tests.txt')

    def test_search_simple(self):
        s = SearchQuery(path='tests/search/resources/search-content-simple-tests.txt')
        results = s.search(r'(ky+o)')

        self.assertListEqual(list(map(lambda x: x[0], results)), [2, 1])

    def test_search_less_simple(self):
        s = SearchQuery(path='tests/search/resources/search-content-simple-tests.txt')
        results = s.search(r'(ky+o) * (uy^{2} + vy + n)')

        self.assertListEqual(list(map(lambda x: x[0], results)), [1, 2])

    def test_search_summation_and_bound_variables(self):
        s = SearchQuery(path='tests/search/resources/search-content-summation-tests.txt')
        results = s.search(r'\sum_{k=0}^{\infty} k')

        self.assertListEqual(list(map(lambda x: x[0], results)), [2, 1])

    def test_remove(self):
        s = SearchQuery(path='tests/search/resources/search-content-simple-tests.txt')
        results = s.search(r'(ky+o) * (uy^{2} + vy + n)')

        self.assertListEqual(list(map(lambda x: x[0], results)), [1, 2])

        s.remove(1)

        results = s.search(r'(ky+o) * (uy^{2} + vy + n)')
        self.assertListEqual(list(map(lambda x: x[0], results)), [2])



from tests import BaseTest

from pymatex.node import Constant, Negate


class ConstantTests(BaseTest.BaseTest):

    def test_read_number(self):
        ast = self.parse('3')
        self.assertEqual(ast, Constant('3'))

    def test_write_number(self):
        ast = self.parse('3')
        self.assertEqual(str(ast), '3')

    def test_read_negative_number(self):
        ast = self.parse('-3')
        self.assertEqual(ast, Negate(Constant('3')))

    def test_write_negative_number(self):
        ast = self.parse('-3')
        self.assertEqual(str(ast), '-(3)')



from enum import Enum, auto
from pymatex.listener import MatexASTVisitor
from pymatex.node.Node import Node


class Func(Enum):
    ABS = auto()
    ACOS = auto()
    ACOSH = auto()
    ACOT = auto()
    ACSC = auto()
    AECOS = auto()
    AEDELTAAMPLITUDE = auto()
    AESIN = auto()
    ASEC = auto()
    ASIN = auto()
    ASINH = auto()
    ATAN = auto()
    ATANH = auto()
    BINOMIAL = auto()
    COS = auto()
    COSH = auto()
    COT = auto()
    CSC = auto()
    ECOS = auto()
    EDELTAAMPLITUDE = auto()
    ESIN = auto()
    EXACTDIVISION = auto()
    FACTORIAL = auto()
    LN = auto()
    LOG = auto()
    SEC = auto()
    SIN = auto()
    SINH = auto()
    SQRT = auto()
    TAN = auto()
    TANH = auto()
    ZETA = auto()


class Function(Node):
    def __init__(self, func: Func, *arguments):
        super().__init__()

        self.func = func
        self.arguments = arguments
        
    def argument(self, i):
        assert i < self.number_of_arguments()
        return self.arguments[i]
        
    def number_of_arguments(self):
        return len(self.arguments)

    def __eq__(self, other):
        return self.func == other.func and self.arguments == other.arguments

    def __str__(self):
        if self.func == Func.ABS:
            return '|{}|'.format(self.argument(0))
        elif self.func == Func.ACSC:
            return '\\arccsc{{{}}}'.format(self.argument(0))
        elif self.func == Func.ACOS:
            return '\\arccos{{{}}}'.format(self.argument(0))
        elif self.func == Func.ACOSH:
            return '\\arccosh{{{}}}'.format(self.argument(0))
        elif self.func == Func.ACOT:
            return '\\arccot{{{}}}'.format(self.argument(0))
        elif self.func == Func.AECOS:
            return '\\arccn{{{}, {}}}'.format(self.argument(0), self.argument(1))
        elif self.func == Func.AEDELTAAMPLITUDE:
            return '\\arcdn{{{}}}'.format(self.argument(0), self.argument(1))
        elif self.func == Func.AESIN:
            return '\\arcsn{{{}}}'.format(self.argument(0), self.argument(1))
        elif self.func == Func.ASEC:
            return '\\arcsec{{{}}}'.format(self.argument(0))
        elif self.func == Func.ASIN:
            return '\\arcsin{{{}}}'.format(self.argument(0))
        elif self.func == Func.ASINH:
            return '\\arcsinh{{{}}}'.format(self.argument(0))
        elif self.func == Func.ATAN:
            return '\\arctan{{{}}}'.format(self.argument(0))
        elif self.func == Func.ATANH:
            return '\\arctanh{{{}}}'.format(self.argument(0))
        elif self.func == Func.BINOMIAL:
            return '\\binom{{{}}}{{{}}}'.format(self.argument(0), self.argument(1))
        elif self.func == Func.COS:
            return '\\cos{{{}}}'.format(self.argument(0))
        elif self.func == Func.COSH:
            return '\\cosh{{{}}}'.format(self.argument(0))
        elif self.func == Func.COT:
            return '\\cot{{{}}}'.format(self.argument(0))
        elif self.func == Func.CSC:
            return '\\csc{{{}}}'.format(self.argument(0))
        elif self.func == Func.ECOS:
            return '\\cn{{{}, {}}}'.format(self.argument(0), self.argument(1))
        elif self.func == Func.EDELTAAMPLITUDE:
            return '\\dn{{{}}}'.format(self.argument(0), self.argument(1))
        elif self.func == Func.ESIN:
            return '\\sn{{{}}}'.format(self.argument(0), self.argument(1))
        elif self.func == Func.EXACTDIVISION:
            return '{}|{}'.format(self.argument(0), self.argument(1))
        elif self.func == Func.FACTORIAL:
            return '({})!'.format(self.argument(0))
        elif self.func == Func.LN:
            return '\\ln{{{}}}'.format(self.argument(0))
        elif self.func == Func.LOG:
            return '\\log{{{}}}'.format(self.argument(0))
        elif self.func == Func.SEC:
            return '\\sec{{{}}}'.format(self.argument(0))
        elif self.func == Func.SIN:
            return '\\sin{{{}}}'.format(self.argument(0))
        elif self.func == Func.SINH:
            return '\\sinh{{{}}}'.format(self.argument(0))
        elif self.func == Func.SQRT:
            return '\\sqrt{{{}}}'.format(self.argument(0))
        elif self.func == Func.TAN:
            return '\\tan{{{}}}'.format(self.argument(0))
        elif self.func == Func.TANH:
            return '\\tanh{{{}}}'.format(self.argument(0))
        elif self.func == Func.ZETA:
            return '\\zeta{{{}}}'.format(self.argument(0))
        else:
            result = r'{}({}'.format(self.func, self.argument(0))
            for i in range(1, self.number_of_arguments()):
                result += ', {}'.format(self.argument(i))
            result += ')'
            return result

    def accept(self, visitor: MatexASTVisitor):
        return visitor.visit_function(self)




# Generated from .\pymatex\grammar\MatexParser.g4 by ANTLR 4.7.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .MatexParser import MatexParser
else:
    from MatexParser import MatexParser

# This class defines a complete listener for a parse tree produced by MatexParser.
class MatexParserListener(ParseTreeListener):

    # Enter a parse tree produced by MatexParser#math.
    def enterMath(self, ctx:MatexParser.MathContext):
        pass

    # Exit a parse tree produced by MatexParser#math.
    def exitMath(self, ctx:MatexParser.MathContext):
        pass


    # Enter a parse tree produced by MatexParser#megaExpr.
    def enterMegaExpr(self, ctx:MatexParser.MegaExprContext):
        pass

    # Exit a parse tree produced by MatexParser#megaExpr.
    def exitMegaExpr(self, ctx:MatexParser.MegaExprContext):
        pass


    # Enter a parse tree produced by MatexParser#specialExpr.
    def enterSpecialExpr(self, ctx:MatexParser.SpecialExprContext):
        pass

    # Exit a parse tree produced by MatexParser#specialExpr.
    def exitSpecialExpr(self, ctx:MatexParser.SpecialExprContext):
        pass


    # Enter a parse tree produced by MatexParser#expr.
    def enterExpr(self, ctx:MatexParser.ExprContext):
        pass

    # Exit a parse tree produced by MatexParser#expr.
    def exitExpr(self, ctx:MatexParser.ExprContext):
        pass


    # Enter a parse tree produced by MatexParser#integralExpr.
    def enterIntegralExpr(self, ctx:MatexParser.IntegralExprContext):
        pass

    # Exit a parse tree produced by MatexParser#integralExpr.
    def exitIntegralExpr(self, ctx:MatexParser.IntegralExprContext):
        pass


    # Enter a parse tree produced by MatexParser#summationExpr.
    def enterSummationExpr(self, ctx:MatexParser.SummationExprContext):
        pass

    # Exit a parse tree produced by MatexParser#summationExpr.
    def exitSummationExpr(self, ctx:MatexParser.SummationExprContext):
        pass


    # Enter a parse tree produced by MatexParser#summationInequalityExpr.
    def enterSummationInequalityExpr(self, ctx:MatexParser.SummationInequalityExprContext):
        pass

    # Exit a parse tree produced by MatexParser#summationInequalityExpr.
    def exitSummationInequalityExpr(self, ctx:MatexParser.SummationInequalityExprContext):
        pass


    # Enter a parse tree produced by MatexParser#productExpr.
    def enterProductExpr(self, ctx:MatexParser.ProductExprContext):
        pass

    # Exit a parse tree produced by MatexParser#productExpr.
    def exitProductExpr(self, ctx:MatexParser.ProductExprContext):
        pass


    # Enter a parse tree produced by MatexParser#productInequalityExpr.
    def enterProductInequalityExpr(self, ctx:MatexParser.ProductInequalityExprContext):
        pass

    # Exit a parse tree produced by MatexParser#productInequalityExpr.
    def exitProductInequalityExpr(self, ctx:MatexParser.ProductInequalityExprContext):
        pass


    # Enter a parse tree produced by MatexParser#tailExpr.
    def enterTailExpr(self, ctx:MatexParser.TailExprContext):
        pass

    # Exit a parse tree produced by MatexParser#tailExpr.
    def exitTailExpr(self, ctx:MatexParser.TailExprContext):
        pass


    # Enter a parse tree produced by MatexParser#funcParams.
    def enterFuncParams(self, ctx:MatexParser.FuncParamsContext):
        pass

    # Exit a parse tree produced by MatexParser#funcParams.
    def exitFuncParams(self, ctx:MatexParser.FuncParamsContext):
        pass


    # Enter a parse tree produced by MatexParser#funcIneqParams.
    def enterFuncIneqParams(self, ctx:MatexParser.FuncIneqParamsContext):
        pass

    # Exit a parse tree produced by MatexParser#funcIneqParams.
    def exitFuncIneqParams(self, ctx:MatexParser.FuncIneqParamsContext):
        pass


    # Enter a parse tree produced by MatexParser#implicitMultiplicationExpr.
    def enterImplicitMultiplicationExpr(self, ctx:MatexParser.ImplicitMultiplicationExprContext):
        pass

    # Exit a parse tree produced by MatexParser#implicitMultiplicationExpr.
    def exitImplicitMultiplicationExpr(self, ctx:MatexParser.ImplicitMultiplicationExprContext):
        pass


    # Enter a parse tree produced by MatexParser#subtractionExpr.
    def enterSubtractionExpr(self, ctx:MatexParser.SubtractionExprContext):
        pass

    # Exit a parse tree produced by MatexParser#subtractionExpr.
    def exitSubtractionExpr(self, ctx:MatexParser.SubtractionExprContext):
        pass


    # Enter a parse tree produced by MatexParser#additionExpr.
    def enterAdditionExpr(self, ctx:MatexParser.AdditionExprContext):
        pass

    # Exit a parse tree produced by MatexParser#additionExpr.
    def exitAdditionExpr(self, ctx:MatexParser.AdditionExprContext):
        pass


    # Enter a parse tree produced by MatexParser#divisionExpr.
    def enterDivisionExpr(self, ctx:MatexParser.DivisionExprContext):
        pass

    # Exit a parse tree produced by MatexParser#divisionExpr.
    def exitDivisionExpr(self, ctx:MatexParser.DivisionExprContext):
        pass


    # Enter a parse tree produced by MatexParser#multiplicationExpr.
    def enterMultiplicationExpr(self, ctx:MatexParser.MultiplicationExprContext):
        pass

    # Exit a parse tree produced by MatexParser#multiplicationExpr.
    def exitMultiplicationExpr(self, ctx:MatexParser.MultiplicationExprContext):
        pass


    # Enter a parse tree produced by MatexParser#fracExpr.
    def enterFracExpr(self, ctx:MatexParser.FracExprContext):
        pass

    # Exit a parse tree produced by MatexParser#fracExpr.
    def exitFracExpr(self, ctx:MatexParser.FracExprContext):
        pass


    # Enter a parse tree produced by MatexParser#powExpr.
    def enterPowExpr(self, ctx:MatexParser.PowExprContext):
        pass

    # Exit a parse tree produced by MatexParser#powExpr.
    def exitPowExpr(self, ctx:MatexParser.PowExprContext):
        pass


    # Enter a parse tree produced by MatexParser#exponentiationExpr.
    def enterExponentiationExpr(self, ctx:MatexParser.ExponentiationExprContext):
        pass

    # Exit a parse tree produced by MatexParser#exponentiationExpr.
    def exitExponentiationExpr(self, ctx:MatexParser.ExponentiationExprContext):
        pass


    # Enter a parse tree produced by MatexParser#signedAtom.
    def enterSignedAtom(self, ctx:MatexParser.SignedAtomContext):
        pass

    # Exit a parse tree produced by MatexParser#signedAtom.
    def exitSignedAtom(self, ctx:MatexParser.SignedAtomContext):
        pass


    # Enter a parse tree produced by MatexParser#negateAtom.
    def enterNegateAtom(self, ctx:MatexParser.NegateAtomContext):
        pass

    # Exit a parse tree produced by MatexParser#negateAtom.
    def exitNegateAtom(self, ctx:MatexParser.NegateAtomContext):
        pass


    # Enter a parse tree produced by MatexParser#localMultiplication.
    def enterLocalMultiplication(self, ctx:MatexParser.LocalMultiplicationContext):
        pass

    # Exit a parse tree produced by MatexParser#localMultiplication.
    def exitLocalMultiplication(self, ctx:MatexParser.LocalMultiplicationContext):
        pass


    # Enter a parse tree produced by MatexParser#atom.
    def enterAtom(self, ctx:MatexParser.AtomContext):
        pass

    # Exit a parse tree produced by MatexParser#atom.
    def exitAtom(self, ctx:MatexParser.AtomContext):
        pass


    # Enter a parse tree produced by MatexParser#variable.
    def enterVariable(self, ctx:MatexParser.VariableContext):
        pass

    # Exit a parse tree produced by MatexParser#variable.
    def exitVariable(self, ctx:MatexParser.VariableContext):
        pass


    # Enter a parse tree produced by MatexParser#indexedVariable.
    def enterIndexedVariable(self, ctx:MatexParser.IndexedVariableContext):
        pass

    # Exit a parse tree produced by MatexParser#indexedVariable.
    def exitIndexedVariable(self, ctx:MatexParser.IndexedVariableContext):
        pass


    # Enter a parse tree produced by MatexParser#constant.
    def enterConstant(self, ctx:MatexParser.ConstantContext):
        pass

    # Exit a parse tree produced by MatexParser#constant.
    def exitConstant(self, ctx:MatexParser.ConstantContext):
        pass


    # Enter a parse tree produced by MatexParser#number.
    def enterNumber(self, ctx:MatexParser.NumberContext):
        pass

    # Exit a parse tree produced by MatexParser#number.
    def exitNumber(self, ctx:MatexParser.NumberContext):
        pass


    # Enter a parse tree produced by MatexParser#absolute.
    def enterAbsolute(self, ctx:MatexParser.AbsoluteContext):
        pass

    # Exit a parse tree produced by MatexParser#absolute.
    def exitAbsolute(self, ctx:MatexParser.AbsoluteContext):
        pass


    # Enter a parse tree produced by MatexParser#exactDivision.
    def enterExactDivision(self, ctx:MatexParser.ExactDivisionContext):
        pass

    # Exit a parse tree produced by MatexParser#exactDivision.
    def exitExactDivision(self, ctx:MatexParser.ExactDivisionContext):
        pass


    # Enter a parse tree produced by MatexParser#factorial.
    def enterFactorial(self, ctx:MatexParser.FactorialContext):
        pass

    # Exit a parse tree produced by MatexParser#factorial.
    def exitFactorial(self, ctx:MatexParser.FactorialContext):
        pass


    # Enter a parse tree produced by MatexParser#binomial.
    def enterBinomial(self, ctx:MatexParser.BinomialContext):
        pass

    # Exit a parse tree produced by MatexParser#binomial.
    def exitBinomial(self, ctx:MatexParser.BinomialContext):
        pass


    # Enter a parse tree produced by MatexParser#func.
    def enterFunc(self, ctx:MatexParser.FuncContext):
        pass

    # Exit a parse tree produced by MatexParser#func.
    def exitFunc(self, ctx:MatexParser.FuncContext):
        pass


    # Enter a parse tree produced by MatexParser#funcname.
    def enterFuncname(self, ctx:MatexParser.FuncnameContext):
        pass

    # Exit a parse tree produced by MatexParser#funcname.
    def exitFuncname(self, ctx:MatexParser.FuncnameContext):
        pass


    # Enter a parse tree produced by MatexParser#bracedMultiExpr.
    def enterBracedMultiExpr(self, ctx:MatexParser.BracedMultiExprContext):
        pass

    # Exit a parse tree produced by MatexParser#bracedMultiExpr.
    def exitBracedMultiExpr(self, ctx:MatexParser.BracedMultiExprContext):
        pass


    # Enter a parse tree produced by MatexParser#parenMultiExpr.
    def enterParenMultiExpr(self, ctx:MatexParser.ParenMultiExprContext):
        pass

    # Exit a parse tree produced by MatexParser#parenMultiExpr.
    def exitParenMultiExpr(self, ctx:MatexParser.ParenMultiExprContext):
        pass


    # Enter a parse tree produced by MatexParser#multiExpr.
    def enterMultiExpr(self, ctx:MatexParser.MultiExprContext):
        pass

    # Exit a parse tree produced by MatexParser#multiExpr.
    def exitMultiExpr(self, ctx:MatexParser.MultiExprContext):
        pass


    # Enter a parse tree produced by MatexParser#bracedExpr.
    def enterBracedExpr(self, ctx:MatexParser.BracedExprContext):
        pass

    # Exit a parse tree produced by MatexParser#bracedExpr.
    def exitBracedExpr(self, ctx:MatexParser.BracedExprContext):
        pass


    # Enter a parse tree produced by MatexParser#brackExpr.
    def enterBrackExpr(self, ctx:MatexParser.BrackExprContext):
        pass

    # Exit a parse tree produced by MatexParser#brackExpr.
    def exitBrackExpr(self, ctx:MatexParser.BrackExprContext):
        pass


    # Enter a parse tree produced by MatexParser#parenExpr.
    def enterParenExpr(self, ctx:MatexParser.ParenExprContext):
        pass

    # Exit a parse tree produced by MatexParser#parenExpr.
    def exitParenExpr(self, ctx:MatexParser.ParenExprContext):
        pass


    # Enter a parse tree produced by MatexParser#subexpr.
    def enterSubexpr(self, ctx:MatexParser.SubexprContext):
        pass

    # Exit a parse tree produced by MatexParser#subexpr.
    def exitSubexpr(self, ctx:MatexParser.SubexprContext):
        pass


    # Enter a parse tree produced by MatexParser#supexpr.
    def enterSupexpr(self, ctx:MatexParser.SupexprContext):
        pass

    # Exit a parse tree produced by MatexParser#supexpr.
    def exitSupexpr(self, ctx:MatexParser.SupexprContext):
        pass


    # Enter a parse tree produced by MatexParser#subeq.
    def enterSubeq(self, ctx:MatexParser.SubeqContext):
        pass

    # Exit a parse tree produced by MatexParser#subeq.
    def exitSubeq(self, ctx:MatexParser.SubeqContext):
        pass


    # Enter a parse tree produced by MatexParser#subIneq.
    def enterSubIneq(self, ctx:MatexParser.SubIneqContext):
        pass

    # Exit a parse tree produced by MatexParser#subIneq.
    def exitSubIneq(self, ctx:MatexParser.SubIneqContext):
        pass


    # Enter a parse tree produced by MatexParser#equality.
    def enterEquality(self, ctx:MatexParser.EqualityContext):
        pass

    # Exit a parse tree produced by MatexParser#equality.
    def exitEquality(self, ctx:MatexParser.EqualityContext):
        pass


    # Enter a parse tree produced by MatexParser#inequality.
    def enterInequality(self, ctx:MatexParser.InequalityContext):
        pass

    # Exit a parse tree produced by MatexParser#inequality.
    def exitInequality(self, ctx:MatexParser.InequalityContext):
        pass





import math
from pymatex.listener import MatexASTVisitor
from pymatex.node.Node import Node


class Constant(Node):

    value: str

    def __init__(self, constant: str):
        super().__init__()

        self.value = None

        if constant == '\\infty':
            self.value = constant
        else:
            _ = float(constant)
            self.value = constant

    def __eq__(self, other):
        return self.value == other.value

    def __str__(self):
        if self.value[0] == '-':
            return '({})'.format(self.value)
        else:
            return self.value

    def accept(self, visitor: MatexASTVisitor):
        return visitor.visit_constant(self)



from pymatex.listener import MatexASTVisitor
from pymatex.node.IterativeFunction import IterativeFunction, IterativeType


class Product(IterativeFunction):

    def __init__(self, variable, start_range, end_range, expression):
        super().__init__(IterativeType.PRODUCT, variable, start_range, end_range, expression)

    def __str__(self):
        return '\\prod_{{{}={}}}^{{{}}}{{{}}}'.format(self.variable, self.start_range, self.end_range,
                                                      self.expression)

    def accept(self, visitor: MatexASTVisitor):
        return visitor.visit_product(self)


class InequalityProduct(Product):

    def __init__(self, start_range, end_range, expression):
        super().__init__(None, start_range, end_range, expression)

    def __str__(self):
        if self.end_range:
            return r'\prod_{{{}}}^{{{}}}{{{}}}'.format(self.start_range, self.end_range, self.expression)
        else:
            return r'\prod_{{{}}}{{{}}}'.format(self.start_range, self.expression)

    def accept(self, visitor: MatexASTVisitor):
        return visitor.visit_product(self)



import csv


class SimilarityMatrix:

    def __init__(self, path=None):

        if path is None:
            path = 'SimilarityMatrix.txt'

        with open(path, 'r') as f:
            csvreader = csv.reader(f, delimiter=';')

            colnames = next(csvreader)[1:]
            self.colnames = [colname.strip() for colname in colnames]
            self.data = [None for _ in range(len(self.colnames) * len(self.colnames))]

            for i, row in enumerate(csvreader):
                row_name = row[0]
                assert row_name in self.colnames
                assert len(row) == len(self.colnames) + 1

                for j, value in enumerate(row[1:]):
                    int_value = int(value)
                    index = i * len(self.colnames) + j
                    self.data[index] = int_value

            assert i == len(self.colnames) - 1

            assert self.is_symmetric()

    def get_similarity_between(self, lhs_node, rhs_node):
        lhs_index = self.colnames.index(lhs_node)
        rhs_index = self.colnames.index(rhs_node)

        index = lhs_index * len(self.colnames) + rhs_index
        return self.data[index]

    def is_symmetric(self):
        for i in range(len(self.colnames)):
            for j in range(len(self.colnames)):
                index_1 = i * len(self.colnames) + j
                index_2 = j * len(self.colnames) + i
                if self.data[index_1] != self.data[index_2]:
                    print("No symmetric value for: {} and {}".format(self.colnames[i], self.colnames[j]))
                    return False
        return True



from tests import BaseTest

from pymatex.node import *


class TaylorMaclaurinTests(BaseTest.BaseTest):

    def test_read_exponential(self):
        ast = self.parse(r'\sum_{n=0}^{\infty} \frac{x^n}{n!}')
        self.assertEqual(ast, Summation(Variable('n'), Constant('0'), Constant('\\infty'),
                                        Division(Exponentiation(Variable('x'), Variable('n')),
                                                 Function(Func.FACTORIAL, Variable('n')))))

    def test_read_logarithm(self):
        ast = self.parse(r'\sum_{n=1}^{\infty} \frac{(-1)^{n+1}}{n} x^{n}')
        self.assertEqual(ast, Summation(Variable('n'), Constant('1'), Constant('\\infty'), Multiplication(
            Division(Exponentiation(Negate(Constant('1')), Addition(Variable('n'), Constant('1'))), Variable('n')),
            Exponentiation(Variable('x'), Variable('n')))))

    def test_read_geometric(self):
        ast = self.parse(r'\sum_{n=0}^{\infty} x^n')
        self.assertEqual(ast, Summation(Variable('n'), Constant('0'), Constant('\\infty'),
                                        Exponentiation(Variable('x'), Variable('n'))))

    def test_read_sin(self):
        ast = self.parse(r'\sum_{n=0}^{\infty} \frac{(-1)^n}{(2n + 1)!} x^{2n + 1}')
        two_times_n_plus_one = Addition(Multiplication(Constant('2'), Variable('n')), Constant('1'))
        self.assertEqual(ast, Summation(Variable('n'), Constant('0'), Constant('\\infty'), Multiplication(
            Division(Exponentiation(Negate(Constant('1')), Variable('n')), Function(Func.FACTORIAL, two_times_n_plus_one)),
            Exponentiation(Variable('x'), two_times_n_plus_one))))

    def test_read_arcsin(self):
        ast = self.parse(r'\sum_{n=0}^{\infty} \frac{(2n)!}{4^n (n!)^2 (2n + 1)} x^{2n + 1}')
        two_times_n = Multiplication(Constant('2'), Variable('n'))
        two_times_n_plus_one = Addition(two_times_n, Constant('1'))
        self.assertEqual(ast, Summation(Variable('n'), Constant('0'), Constant('\\infty'), Multiplication(
            Division(Function(Func.FACTORIAL, two_times_n),
                     Multiplication(Multiplication(Exponentiation(Constant('4'), Variable('n')),
                                    Exponentiation(Function(Func.FACTORIAL, Variable('n')), Constant('2'))),
                                    two_times_n_plus_one)),
            Exponentiation(Variable('x'), two_times_n_plus_one))))

    def test_read_tan(self):
        ast = self.parse(r'\sum_{n=1}^{\infty} |B_{2n}| \frac{4^n (4^n - 1)}{(2n)!} x^{2n - 1}')
        two_times_n = Multiplication(Constant('2'), Variable('n'))
        four_exponent_n = Exponentiation(Constant('4'), Variable('n'))

        self.assertEqual(ast, Summation(Variable('n'), Constant('1'), Constant('\\infty'), Multiplication(
            Multiplication(Function(Func.ABS, IndexedVariable('B', two_times_n)), Division(
                Multiplication(four_exponent_n, Subtraction(four_exponent_n, Constant('1'))),
                Function(Func.FACTORIAL, Multiplication(Constant('2'), Variable('n')))
            )), Exponentiation(Variable('x'), Subtraction(two_times_n, Constant('1'))))))

    def test_read_exponent_alpha_arcsin(self):
        ast = self.parse(r'\sum_{n=0}^{\infty }{\frac {\prod _{k=0}^{n-1}(4k^{2}+\alpha ^{2})}{(2n)!}z^{2n}}+\sum _{n=0}^{\infty }{\frac {\alpha \prod _{k=0}^{n-1}[(2k+1)^{2}+\alpha ^{2}]}{(2n+1)!}z^{2n+1}}')

        a = Variable('\\alpha')
        k = Variable('k')
        n = Variable('n')
        o = Constant('1')
        t = Constant('2')
        z = Constant('0')
        infinity = Constant('\\infty')

        first_summation = Summation(n, z, infinity, Multiplication(
            Division(Product(k, z, Subtraction(n, o), Addition(
                Multiplication(Constant('4'), Exponentiation(k, t)),
                Exponentiation(a, t))),
                     Function(Func.FACTORIAL, Multiplication(t, n))),
            Exponentiation(Variable('z'), Multiplication(t, n))))

        second_summation = Summation(n, z, infinity, Multiplication(
            Division(Multiplication(a, Product(k, z, Subtraction(n, o), Addition(
                Exponentiation(Addition(Multiplication(t, k), o), t),
                Exponentiation(a, t)))),
                     Function(Func.FACTORIAL, Addition(Multiplication(t, n), o))),
            Exponentiation(Variable('z'), Addition(Multiplication(t, n), o))))

        self.assertEqual(ast, Addition(first_summation, second_summation))



from antlr4 import CommonTokenStream, InputStream, ParseTreeWalker
from pymatex.error import ExceptionErrorListener
from pymatex.grammar import MatexLexer, MatexParser, MatexParserListener
from pymatex.listener import MatexASTVisitor
from pymatex.node import *


class MatexAST(MatexParserListener):

    def __init__(self):
        super().__init__()
        self.__stack = []
        self.__head = None

    def __eq__(self, other):
        assert (self.__head is not None)
        return self.__head == other

    def __str__(self):
        assert (self.__head is not None)
        return str(self.__head)

    def accept(self, visitor: MatexASTVisitor):
        assert (self.__head is not None)
        return self.__head.accept(visitor)

    @staticmethod
    def get_tree(latex: str):
        lexer = MatexLexer(InputStream(latex))
        stream = CommonTokenStream(lexer)
        parser = MatexParser(stream)
        parser._listeners = [ExceptionErrorListener()]
        return parser.math()

    @staticmethod
    def parse(latex: str):
        tree = MatexAST.get_tree(latex)
        ast = MatexAST()
        walker = ParseTreeWalker()
        walker.walk(ast, tree)
        return ast

    def exitAbsolute(self, ctx: MatexParser.AbsoluteContext):
        self.push(Function(Func.ABS, self.pop()))

    def exitAdditionExpr(self, ctx: MatexParser.AdditionExprContext):
        if ctx.getChildCount() < 2:
            return

        rhs = self.pop()
        lhs = self.pop()
        self.push(Addition(lhs, rhs))

    def exitBinomial(self, ctx: MatexParser.BinomialContext):
        if ctx.getChildCount() < 2:
            return

        rhs = self.pop()
        lhs = self.pop()
        self.push(Function(Func.BINOMIAL, lhs, rhs))

    def exitConstant(self, ctx: MatexParser.ConstantContext):
        if ctx.INFINITY():
            self.push(Constant(ctx.INFINITY().getText()))

    def exitDivisionExpr(self, ctx: MatexParser.DivisionExprContext):
        if ctx.getChildCount() < 2:
            return

        rhs = self.pop()
        lhs = self.pop()
        self.push(Division(lhs, rhs))

    def exitExactDivision(self, ctx: MatexParser.ExactDivisionContext):
        self.push(Function(Func.EXACTDIVISION, Variable(ctx.VARIABLE(0).getText()), Variable(ctx.VARIABLE(1).getText())))

    def exitExponentiationExpr(self, ctx: MatexParser.ExponentiationExprContext):
        rhs = self.pop()
        lhs = self.pop()

        # Fix the case where 2a^2 is interpreted as (2 * a)^2
        if isinstance(lhs, Multiplication):
            if isinstance(lhs.lhs, Constant) and isinstance(lhs.rhs, Variable):
                self.push(Multiplication(lhs.lhs, Exponentiation(lhs.rhs, rhs)))
                return

        self.push(Exponentiation(lhs, rhs))

    def exitFactorial(self, ctx: MatexParser.FactorialContext):
        expression = self.pop()
        self.push(Function(Func.FACTORIAL, expression))

    def exitFracExpr(self, ctx: MatexParser.FracExprContext):
        rhs = self.pop()
        lhs = self.pop()
        self.push(Division(lhs, rhs))

    def exitFunc(self, ctx: MatexParser.FuncContext):
        funcname: MatexParser.FuncnameContext = ctx.funcname()

        if funcname is None:
            greek_letter = ctx.GREEKFUNCTIONPAREN()
            if greek_letter is None:
                greek_letter = ctx.GREEKFUNCTIONBRACE()
            if greek_letter is None:
                greek_letter = ctx.LETTERFUNCTIONBRACE()
            if greek_letter is None:
                greek_letter = ctx.LETTERFUNCTIONPAREN()

            letter = greek_letter.getText()[:-1]

            arguments = ctx.bracedMultiExpr()
            if not arguments:
                arguments = ctx.parenMultiExpr()
            if not arguments:
                arguments = ctx.multiExpr()

            expressions = []
            for _ in range((arguments.getChildCount() + 1) // 2):
                expressions.insert(0, self.pop())

            if letter == r'\zeta':
                return self.push(Function(Func.ZETA, *expressions))

            return self.push(Function(letter, *expressions))

        argument_0 = self.pop()
        if funcname.FUNC_ARCCOS():
            self.push(Function(Func.ACOS, argument_0))
        elif funcname.FUNC_ARCCOSH():
            self.push(Function(Func.ACOSH, argument_0))
        elif funcname.FUNC_ARCCOT():
            self.push(Function(Func.ACOST, argument_0))
        elif funcname.FUNC_ARCCSC():
            self.push(Function(Func.ACSC, argument_0))
        elif funcname.FUNC_ARCSEC():
            self.push(Function(Func.ASEC, argument_0))
        elif funcname.FUNC_ARCECOS():
            self.push(Function(Func.AECOS, self.pop(), argument_0))
        elif funcname.FUNC_ARCEDELTAAMPLITUDE():
            self.push(Function(Func.AEDELTAAMPLITUDE, self.pop(), argument_0))
        elif funcname.FUNC_ARCESIN():
            self.push(Function(Func.AESIN, self.pop(), argument_0))
        elif funcname.FUNC_ARCSIN():
            self.push(Function(Func.ASIN, argument_0))
        elif funcname.FUNC_ARCSINH():
            self.push(Function(Func.ASINH, argument_0))
        elif funcname.FUNC_ARCTAN():
            self.push(Function(Func.ATAN, argument_0))
        elif funcname.FUNC_ARCTANH():
            self.push(Function(Func.ATANH, argument_0))
        elif funcname.FUNC_COS():
            self.push(Function(Func.COS, argument_0))
        elif funcname.FUNC_COSH():
            self.push(Function(Func.COSH, argument_0))
        elif funcname.FUNC_COT():
            self.push(Function(Func.COT, argument_0))
        elif funcname.FUNC_CSC():
            self.push(Function(Func.CSC, argument_0))
        elif funcname.FUNC_ECOS():
            self.push(Function(Func.ECOS, self.pop(), argument_0))
        elif funcname.FUNC_EDELTAAMPLITUDE():
            self.push(Function(Func.EDELTAAMPLITUDE, self.pop(), argument_0))
        elif funcname.FUNC_ESIN():
            self.push(Function(Func.ESIN, self.pop(), argument_0))
        elif funcname.FUNC_LN():
            self.push(Function(Func.LN, argument_0))
        elif funcname.FUNC_LOG():
            self.push(Function(Func.LOG, argument_0))
        elif funcname.FUNC_SEC():
            self.push(Function(Func.SEC, argument_0))
        elif funcname.FUNC_SIN():
            self.push(Function(Func.SIN, argument_0))
        elif funcname.FUNC_SINH():
            self.push(Function(Func.SINH, argument_0))
        elif funcname.FUNC_SQRT():
            self.push(Function(Func.SQRT, argument_0))
        elif funcname.FUNC_TAN():
            self.push(Function(Func.TAN, argument_0))
        elif funcname.FUNC_ARCTAN():
            self.push(Function(Func.ATAN, argument_0))

    def exitIndexedVariable(self, ctx: MatexParser.IndexedVariableContext):
        expression = self.pop()
        self.push(IndexedVariable(ctx.VARIABLE().getText(), expression))

    def exitInequality(self, ctx: MatexParser.InequalityContext):
        rhs = self.pop()
        lhs = self.pop()
        self.push(Inequality(lhs, rhs, ctx.INEQUALITIES().getText()))

    def exitIntegralExpr(self, ctx: MatexParser.IntegralExprContext):
        expression = self.pop()
        end_range = self.pop()
        start_range = self.pop()

        variable = None
        if ctx.DERIVATIVE():
            variable = Variable(ctx.DERIVATIVE().getText()[1:])

        self.push(Integral(variable, start_range, end_range, expression))

    def exitLocalMultiplication(self, ctx: MatexParser.LocalMultiplicationContext):
        last = None
        variables = None

        if ctx.MIXNUMBER():
            mixed_number = ctx.MIXNUMBER().getText()

            import re
            for it in re.finditer(r'[-+]?\d*\.?\d+([eE][-+]?\d+)?', mixed_number):
                span = it.span(0)
                number = mixed_number[span[0]:span[1]]
                last = Constant(number)
                variables = mixed_number[span[1]:]
        elif ctx.WORD():
            variables = ctx.WORD().getText()
        else:
            variables = ctx.DERIVATIVE().getText()

        for variable in variables:
            if last is None:
                last = Variable(variable)
            else:
                last = Multiplication(last, Variable(variable))
        self.push(last)

    def exitNegateAtom(self, ctx: MatexParser.NegateAtomContext):
        value = self.pop()
        self.push(Negate(value))

    def exitNumber(self, ctx: MatexParser.NumberContext):
        self.push(Constant(ctx.NUMBER().getText()))

    def exitMath(self, ctx: MatexParser.MathContext):
        assert (len(self.__stack) == 1)
        self.__head = self.__stack[0]

    def exitMegaExpr(self, ctx: MatexParser.MegaExprContext):
        if ctx.getChildCount() < 2:
            return

        rhs = self.pop()
        lhs = self.pop()

        if ctx.PLUS():
            self.push(Addition(lhs, rhs))
        elif ctx.MUL():
            self.push(Multiplication(lhs, rhs))
        elif ctx.MINUS():
            self.push(Subtraction(lhs, rhs))
        else:
            self.push(Multiplication(lhs, rhs))

    def exitImplicitMultiplicationExpr(self, ctx: MatexParser.ImplicitMultiplicationExprContext):
        if ctx.getChildCount() < 2:
            return

        rhs = self.pop()
        lhs = self.pop()

        self.push(Multiplication(lhs, rhs))

    def exitMultiplicationExpr(self, ctx: MatexParser.MultiplicationExprContext):
        if ctx.getChildCount() < 2:
            return

        rhs = self.pop()
        lhs = self.pop()

        self.push(Multiplication(lhs, rhs))

    def exitProductExpr(self, ctx: MatexParser.ProductExprContext):
        expression = self.pop()
        end_range = self.pop()
        start_range = self.pop()
        variable = self.pop()
        self.push(Product(variable, start_range, end_range, expression))

    def exitProductInequalityExpr(self, ctx: MatexParser.SummationInequalityExprContext):
        expression = self.pop()
        end_range = None
        if ctx.funcIneqParams().supexpr():
            end_range = self.pop()
        start_range = self.pop()

        self.push(InequalityProduct(start_range, end_range, expression))

    def exitSubtractionExpr(self, ctx: MatexParser.SubtractionExprContext):
        if ctx.getChildCount() < 2:
            return

        rhs = self.pop()
        lhs = self.pop()
        self.push(Subtraction(lhs, rhs))

    def exitSummationExpr(self, ctx: MatexParser.SummationExprContext):
        expression = self.pop()
        end_range = self.pop()
        start_range = self.pop()
        variable = self.pop()

        self.push(Summation(variable, start_range, end_range, expression))

    def exitSummationInequalityExpr(self, ctx: MatexParser.SummationInequalityExprContext):
        expression = self.pop()
        end_range = None
        if ctx.funcIneqParams().supexpr():
            end_range = self.pop()
        start_range = self.pop()

        self.push(InequalitySummation(start_range, end_range, expression))

    def exitVariable(self, ctx: MatexParser.VariableContext):
        if ctx.VARIABLE():
            self.push(Variable(ctx.VARIABLE().getText()))
        else:
            self.push(Variable(ctx.GREEKLETTER().getText()))

    def pop(self):
        assert (len(self.__stack) >= 1)
        return self.__stack.pop()

    def push(self, value):
        self.__stack.append(value)



from pymatex.listener import MatexASTVisitor
from pymatex.node.Node import Node


class Negate(Node):

    def __init__(self, node):
        super().__init__()

        self.node = node

    def __eq__(self, other):
        return self.node == other.node

    def __str__(self):
        return '-({})'.format(self.node)

    def accept(self, visitor: MatexASTVisitor):
        return visitor.visit_negate(self)



from pymatex.listener import MatexASTVisitor
from pymatex.node.BinaryOperator import BinaryOperator


class Multiplication(BinaryOperator):

    def __init__(self, lhs, rhs):
        super().__init__(lhs, rhs, '*', None)

    def accept(self, visitor: MatexASTVisitor):
        return visitor.visit_multiplication(self)
